# -*- coding: utf-8 -*-
"""
Created on Sat Dec 16 10:16:50 2023
Engine to train the surrogate with

@author: Rebecca Kohlhaas
"""
import copy
import h5py
import joblib
import numpy as np
import os

from .inputs import Input
from .exp_designs import ExpDesigns
from .exp_designs_ import ExpDesigns_
from .surrogate_models import MetaModel
from bayesvalidrox.post_processing.post_processing import PostProcessing

class Engine():
    
    
    def __init__(self, MetaMod, Model, ExpDes):
        self.MetaModel = MetaMod
        self.Model = Model
        self.ExpDesign = ExpDes
        
    def start_engine(self) -> None:
        """
        Do all the preparations that need to be run before the actual training

        Returns
        -------
        None

        """
        self.out_names = self.Model.Output.names
        self.MetaModel.out_names = self.out_names
        
        
    def train_normal(self, parallel = False, verbose = False, save = False) -> None:
        """
        Trains surrogate on static samples only.
        Samples are taken from the experimental design and the specified 
        model is run on them.
        Alternatively the samples can be read in from a provided hdf5 file.
        

        Returns
        -------
        None

        """
            
        ExpDesign = self.ExpDesign
        MetaModel = self.MetaModel
        # Read ExpDesign (training and targets) from the provided hdf5
        if ExpDesign.hdf5_file is not None:
            # TODO: need to run 'generate_ED' as well after this or not?
            ExpDesign.read_from_file(self.out_names)
        else:
            # Check if an old hdf5 file exists: if yes, rename it
            hdf5file = f'ExpDesign_{self.Model.name}.hdf5'
            if os.path.exists(hdf5file):
                os.rename(hdf5file, 'old_'+hdf5file)

        # ---- Prepare X samples ----
        # For training the surrogate use ExpDesign.X_tr, ExpDesign.X is for the model to run on 
        ExpDesign.generate_ED(ExpDesign.n_init_samples,
                                              transform=True,
                                              max_pce_deg=np.max(MetaModel.pce_deg))
        
        # ---- Run simulations at X ----
        if not hasattr(ExpDesign, 'Y') or ExpDesign.Y is None:
            print('\n Now the forward model needs to be run!\n')
            ED_Y, up_ED_X = self.Model.run_model_parallel(ExpDesign.X)
            ExpDesign.Y = ED_Y
        else:
            # Check if a dict has been passed.
            if not type(ExpDesign.Y) is dict:
                raise Exception('Please provide either a dictionary or a hdf5'
                                'file to ExpDesign.hdf5_file argument.')
                
        # Separate output dict and x-values
        if 'x_values' in ExpDesign.Y:
            ExpDesign.x_values = ExpDesign.Y['x_values']
            del ExpDesign.Y['x_values']
        
        
        # Build and train the MetaModel on the static samples
        MetaModel.CollocationPoints = ExpDesign.X_tr
        MetaModel.build_metamodel(ExpDesign.n_init_samples)
        
        # Fit the surrogate
        MetaModel.fit(ExpDesign.X_tr, ExpDesign.Y, parallel, verbose)
        
        # TODO: this just a fix for the evals - remove later
        MetaModel.ExpDesign = ExpDesign
        
        # Save what there is to save
        if save:
            # Save surrogate
            with open(f'surrogates/surrogate_{self.Model.name}.pk1', 'wb') as output:
                joblib.dump(MetaModel, output, 2)
                    
            # Zip the model run directories
            if self.Model.link_type.lower() == 'pylink' and\
               self.ExpDesign.sampling_method.lower() != 'user':
                self.Model.zip_subdirs(self.Model.name, f'{self.Model.name}_')
                
            
    def train_sequential(self, parallel = False, verbose = False) -> None:
        """
        Train the surrogate in a sequential manner.
        First build and train evereything on the static samples, then iterate
        choosing more samples and refitting the surrogate on them.

        Returns
        -------
        None

        """
        self.train_normal(parallel, verbose)
        
        
    
    # -------------------------------------------------------------------------
    def eval_metamodel(self, samples=None, nsamples=None,
                       sampling_method='random', return_samples=False):
        """
        Evaluates meta-model at the requested samples. One can also generate
        nsamples.

        Parameters
        ----------
        samples : array of shape (n_samples, n_params), optional
            Samples to evaluate meta-model at. The default is None.
        nsamples : int, optional
            Number of samples to generate, if no `samples` is provided. The
            default is None.
        sampling_method : str, optional
            Type of sampling, if no `samples` is provided. The default is
            'random'.
        return_samples : bool, optional
            Retun samples, if no `samples` is provided. The default is False.

        Returns
        -------
        mean_pred : dict
            Mean of the predictions.
        std_pred : dict
            Standard deviatioon of the predictions.
        """
        # Generate or transform (if need be) samples
        if samples is None:
            # Generate
            samples = self.ExpDesign.generate_samples(
                nsamples,
                sampling_method
                )

        # Transform samples to the independent space
        samples = self.ExpDesign.transform(
            samples,
            method='user'
            )
        # print(samples)

        # Compute univariate bases for the given samples
        if self.meta_model_type.lower() != 'gpe':
            univ_p_val = self.univ_basis_vals(
                samples,
                n_max=np.max(self.pce_deg)
                )

        mean_pred_b = {}
        std_pred_b = {}
        # Loop over bootstrap iterations
        for b_i in range(self.n_bootstrap_itrs):

            # Extract model dictionary
            if self.meta_model_type.lower() == 'gpe':
                model_dict = self.gp_poly[f'b_{b_i+1}']
            else:
                model_dict = self.coeffs_dict[f'b_{b_i+1}']

            # Loop over outputs
            mean_pred = {}
            std_pred = {}
            for output, values in model_dict.items():

                mean = np.empty((len(samples), len(values)))
                std = np.empty((len(samples), len(values)))
                idx = 0
                for in_key, InIdxValues in values.items():

                    # Perdiction with GPE
                    if self.meta_model_type.lower() == 'gpe':
                        X_T = self.x_scaler[f'b_{b_i+1}'][output].transform(samples)
                        gp = self.gp_poly[f'b_{b_i+1}'][output][in_key]
                        y_mean, y_std = gp.predict(X_T, return_std=True)

                    else:
                        # Perdiction with PCE
                        # Assemble Psi matrix
                        basis = self.basis_dict[f'b_{b_i+1}'][output][in_key]
                        psi = self.create_psi(basis, univ_p_val)

                        # Perdiction
                        if self.bootstrap_method != 'fast' or b_i == 0:
                            # with error bar, i.e. use clf_poly
                            clf_poly = self.clf_poly[f'b_{b_i+1}'][output][in_key]
                            try:
                                y_mean, y_std = clf_poly.predict(
                                    psi, return_std=True
                                    )
                            except TypeError:
                                y_mean = clf_poly.predict(psi)
                                y_std = np.zeros_like(y_mean)
                        else:
                            # without error bar
                            coeffs = self.coeffs_dict[f'b_{b_i+1}'][output][in_key]
                            y_mean = np.dot(psi, coeffs)
                            y_std = np.zeros_like(y_mean)

                    mean[:, idx] = y_mean
                    std[:, idx] = y_std
                    idx += 1

                # Save predictions for each output
                if self.dim_red_method.lower() == 'pca':
                    PCA = self.pca[f'b_{b_i+1}'][output]
                    mean_pred[output] = PCA.inverse_transform(mean)
                    std_pred[output] = np.zeros(mean.shape)
                else:
                    mean_pred[output] = mean
                    std_pred[output] = std

            # Save predictions for each bootstrap iteration
            mean_pred_b[b_i] = mean_pred
            std_pred_b[b_i] = std_pred

        # Change the order of nesting
        mean_pred_all = {}
        for i in sorted(mean_pred_b):
            for k, v in mean_pred_b[i].items():
                if k not in mean_pred_all:
                    mean_pred_all[k] = [None] * len(mean_pred_b)
                mean_pred_all[k][i] = v

        # Compute the moments of predictions over the predictions
        for output in self.out_names:
            # Only use bootstraps with finite values
            finite_rows = np.isfinite(
                mean_pred_all[output]).all(axis=2).all(axis=1)
            outs = np.asarray(mean_pred_all[output])[finite_rows]
            # Compute mean
            mean_pred[output] = np.mean(outs, axis=0)
            # Compute standard deviation
            if self.n_bootstrap_itrs > 1:
                std_pred[output] = np.std(outs, axis=0)
            else:
                std_pred[output] = std_pred_b[b_i][output]

        if return_samples:
            return mean_pred, std_pred, samples
        else:
            return mean_pred, std_pred